from typing import Union, Dict, Any


class Action:
    """
    DTO representing an action in a simulation environment
    """

    def __init__(self, id: Union[int, float], descr: str):
        """
        Initializes the DTO

        :param id: the id of the action
        :param descr: the description of the action
        """
        self.id = id
        self.descr = descr

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "Action":
        """
        Converts a dict representation of the DTO into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = Action(
            id=d["id"], descr=d["descr"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the DTO
        """
        d = {}
        d["id"] = self.id
        d["descr"] = self.descr
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"id:{self.id}, descr:{self.descr}"



from typing import Union, List, Dict
from abc import ABC, abstractmethod
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType


class Policy(ABC):

    def __init__(self, agent_type: AgentType, player_type: PlayerType):
        self.agent_type = agent_type
        self.player_type = player_type

    @abstractmethod
    def action(self, o: Union[List[Union[int, float]], int, float]) -> Union[int, float]:
        pass

    @abstractmethod
    def to_dict(self) -> Dict:
        pass

    @abstractmethod
    def stage_policy(self, o: Union[List[Union[int, float]], int, float]) -> List[List[float]]:
        pass

    @staticmethod
    @abstractmethod
    def from_dict(d: Dict) -> "Policy":
        pass

    @abstractmethod
    def probability(self, o: Union[List[Union[int, float]], int, float], a: int) -> float:
        pass

    @abstractmethod
    def copy(self) -> "Policy":
        pass



from typing import Set
from csle_common.dao.domain_randomization.randomization_space import RandomizationSpace
from csle_common.dao.emulation_config.node_vulnerability_config import NodeVulnerabilityConfig
from csle_common.dao.emulation_config.node_users_config import NodeUsersConfig
from csle_common.dao.emulation_config.node_flags_config import NodeFlagsConfig


class NodeRandomizerConfig:
    """
    DTO representing domain randomization configuration of a single node
    """

    def __init__(self, ip: str, reachable: Set[str], id: int, users_config: NodeUsersConfig,
                 flags_config: NodeFlagsConfig, vulns_config: NodeVulnerabilityConfig,
                 r_space: RandomizationSpace,
                 router: bool = False, agent: bool = False, gateway: bool = False):
        """
        Creates a config DTO for randomizing a node in a csle env

        :param ip: the ip of the node
        :param reachable: the list of reachable nodes
        :param id: the id of the node
        :param users_config: the configuration of the users
        :param flags_config: the configuration of the flags
        :param vulns_config: the configuration of the vulnerabilities
        :param r_space: the randomization space
        :param router: the router
        :param agent: the agent
        :param gateway: the gateway
        :return: the randomized node
        """
        self.ip = ip
        self.reachable = reachable
        self.id = id
        self.users_config = users_config
        self.flags_config = flags_config
        self.vulns_config = vulns_config
        self.r_space = r_space
        self.router = router
        self.agent = agent
        self.gateway = gateway

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip: {self.ip}, reachable: {self.reachable}, id: {self.id}, users_config: {self.users_config}, " \
               f"flags_config: {self.flags_config}, vulns_config: {self.vulns_config}, r_space: {str(self.r_space)}," \
               f"router: {self.router}, agent: {self.agent}, gateway: {self.gateway}"



from typing import List, Dict, Any


class TransitionOperatorConfig:
    """
    DTO representing the transition operator definition of a simulation
    """

    def __init__(self, transition_tensor: List):
        """
        Initializes the DTO

        :param transition_tensor: the transition tensor of the simulation
        """
        self.transition_tensor = transition_tensor

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "TransitionOperatorConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = TransitionOperatorConfig(
            transition_tensor=d["transition_tensor"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the object
        """
        d = {}
        d["transition_tensor"] = self.transition_tensor
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"transition tensor: {self.transition_tensor}"



from typing import List
from csle_common.dao.emulation_action_result.nikto_vuln import NiktoVuln


class NiktoScanResult:
    """
    DTO representing the result of a NiktoScan
    """

    def __init__(self, vulnerabilities: List[NiktoVuln], port: int, ip: str, sitename: str):
        """
        Initializes the DTO

        :param vulnerabilities: the list of found vulnerabilities from the scan
        :param port: the port of the scan
        :param ip: the ip of the scan
        :param sitename: the sitename of the scan
        """
        self.vulnerabilities = vulnerabilities
        self.port = port
        self.ip = ip
        self.sitename = sitename

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip:{self.ip}, port:{self.port}, sitename:{self.sitename}, " \
               f"vulnerabilities:{' '.join(list(map(lambda x: str(x), self.vulnerabilities)))}"



from typing import List
import copy
from csle_common.dao.emulation_action_result.nmap_host_status import NmapHostStatus
from csle_common.dao.emulation_action_result.nmap_port import NmapPort
from csle_common.dao.emulation_action_result.nmap_os import NmapOs
from csle_common.dao.emulation_action_result.nmap_vuln import NmapVuln
from csle_common.dao.emulation_action_result.nmap_brute_credentials import NmapBruteCredentials
from csle_common.dao.emulation_action_result.nmap_trace import NmapTrace


class NmapHostResult:
    """
    A DTO representing a host found with an NMAP scan
    """

    def __init__(self, status: NmapHostStatus = NmapHostStatus.DOWN, ips: List[str] = None,
                 mac_addr: str = None, hostnames: List[str] = None,
                 ports: List[NmapPort] = None, os: NmapOs = None, os_matches: List[NmapOs] = None,
                 vulnerabilities: List[NmapVuln] = None, credentials: List[NmapBruteCredentials] = None,
                 trace: NmapTrace = None):
        """
        Initializes the DTO

        :param status: the status of the host
        :param ips: the ip address of the host
        :param mac_addr: the mac address of the host
        :param hostnames: the hostnames of the host
        :param ports: the ports of the host
        :param os: the operating system of the host
        :param os_matches: the matched operating system of the host
        :param vulnerabilities: the vulnerabilities of the host
        :param credentials: the credentials of the host
        :param trace: the trace of the host
        """
        self.status = status
        self.ips = ips
        if self.ips is None:
            self.ips = []
        self.mac_addr = mac_addr
        self.hostnames = hostnames
        self.ports = ports
        self.os = os
        self.os_matches = os_matches
        self.vulnerabilities = vulnerabilities
        self.credentials = credentials
        self.trace = trace

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"status:{self.status}, ip_addr:{self.ips}, mac_addr:{self.mac_addr}, " \
               f"hostnames:{' '.join(self.hostnames)}, " \
               f"ports:{' '.join(list(map(lambda x: str(x), self.ports)))}, os:{self.os}, " \
               f"os_matches:{' '.join(list(map(lambda x: str(x), self.os_matches)))}, " \
               f"vulnerabilities:{' '.join(list(map(lambda x: str(x), self.vulnerabilities)))}, " \
               f"credentials:{' '.join(list(map(lambda x: str(x), self.credentials)))}, " \
               f"trace:{self.trace}"

    def copy(self) -> "NmapHostResult":
        """
        :return: a copy of the object
        """
        return copy.deepcopy(self)

    def ips_match(self, ips: List[str]) -> bool:
        """
        Checks if a list of ips overlap with the ips of this host

        :param ips: the list of ips to check
        :return:  True if they match, False otherwise
        """
        for ip in self.ips:
            if ip in ips:
                return True
        return False



from typing import List
from csle_common.dao.emulation_action_result.nmap_host_result import NmapHostResult


class NmapScanResult:
    """
    DTO representing the result of a NMAP Scan
    """

    def __init__(self, hosts: List[NmapHostResult], ips: List[str]):
        """
        Initializes the DTO

        :param hosts: the list of NMAP hosts
        :param ips: the ips of the host
        """
        self.hosts = hosts
        self.ips = ips
        self.reachable = []
        for h in self.hosts:
            self.reachable = self.reachable + h.ips

    def __str__(self) -> str:
        """
        :return:  a string representation of the objectp
        """
        return "hosts:{}".format(" ".join(list(map(lambda x: str(x), self.hosts))))

    def copy(self) -> "NmapScanResult":
        """
        :return: a copy of the object
        """
        hosts = []
        for host in self.hosts:
            hosts.append(host.copy())
        return NmapScanResult(hosts=hosts, ips=self.ips)



from typing import Dict, Any, List


class RewardFunctionConfig:
    """
    DTO containing the reward tensor of a simulation
    """

    def __init__(self, reward_tensor: List):
        """
        Initalizes the DTO

        :param reward_tensor: the reward tensor of the simulation
        """
        self.reward_tensor = reward_tensor

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "RewardFunctionConfig":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = RewardFunctionConfig(
            reward_tensor=d["reward_tensor"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation  of the object
        """
        d = {}
        d["reward_tensor"] = self.reward_tensor
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"reward_tensor:{self.reward_tensor}"



from typing import List, Dict, Any
from csle_common.dao.emulation_config.node_resources_config import NodeResourcesConfig


class ResourcesConfig:
    """
    A DTO representing the resources assigned to the containers in an emulation environment
    """

    def __init__(self, node_resources_configurations: List[NodeResourcesConfig]):
        """
        Initializes the DTO

        :param node_resources_configurations: the list resource configurations for each node
        """
        self.node_resources_configurations = node_resources_configurations

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ResourcesConfig":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ResourcesConfig(
            node_resources_configurations=list(map(lambda x: NodeResourcesConfig.from_dict(x),
                                                   d["node_resources_configurations"]))
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the object
        """
        d = {}
        d["node_resources_configurations"] = list(map(lambda x: x.to_dict(), self.node_resources_configurations))
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return ",".join(list(map(lambda x: str(x), self.node_resources_configurations)))

    def to_json_str(self) -> str:
        """
        Converts the DTO into a json string

        :return: the json string representation of the DTO
        """
        import json
        json_str = json.dumps(self.to_dict(), indent=4, sort_keys=True)
        return json_str

    def to_json_file(self, json_file_path: str) -> None:
        """
        Saves the DTO to a json file

        :param json_file_path: the json file path to save  the DTO to
        :return: None
        """
        import io
        json_str = self.to_json_str()
        with io.open(json_file_path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    @staticmethod
    def from_json_file(json_file_path: str) -> "ResourcesConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ResourcesConfig.from_dict(json.loads(json_str))

    def copy(self) -> "ResourcesConfig":
        """
        :return: a copy of the DTO
        """
        return ResourcesConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "ResourcesConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.node_resources_configurations = list(map(lambda x: x.create_execution_config(
            ip_first_octet=ip_first_octet), self.node_resources_configurations))
        return config



from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.credential import Credential


class NmapBruteCredentials:
    """
    A DTO representing credentials found with NMAP Brute-Force Scans
    """

    def __init__(self, username: str, pw: str, state: str, port: int, protocol: TransportProtocol, service: str):
        """
        Initializes the DTO

        :param username: the username of the credential
        :param pw: the pw of the credential
        :param state: the state of the credential
        :param port: the port of the scan
        :param protocol: the protocol of the scan
        :param service: the service of the credential
        """
        self.username = username
        self.pw = pw
        self.state = state
        self.port = port
        self.protocol = protocol
        self.service = service

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return "username:{},pw:{},state:{},port:{},protocol:{},service:{}".format(
            self.username, self.pw, self.state, self.port, self.protocol, self.service)

    def __hash__(self) -> int:
        """
        :return: a hash of the object
        """
        return hash(self.username) + 31 * hash(self.pw)

    def __eq__(self, other) -> bool:
        """
        Checks equality with another object

        :param other: the object to compare with
        :return: True if equal otherwise False
        """
        return (self.username == other.username and
                self.pw == other.pw)

    def to_obs(self) -> Credential:
        """
        Converts the Object into a Credential Object

        :return: the created Credential Object
        """
        return Credential(username=self.username, pw=self.pw, port=int(self.port), service=self.service,
                          protocol=self.protocol)



from typing import Dict, Any


class SnortIDSManagerConfig:
    """
    Represents the configuration of the Snort IDS managers in a CSLE emulation
    """

    def __init__(self, snort_ids_manager_log_file: str, snort_ids_manager_log_dir: str,
                 snort_ids_manager_max_workers: int,
                 time_step_len_seconds: int = 15, snort_ids_manager_port: int = 50048, version: str = "0.0.1") -> None:
        """
        Initializes the DTO

        :param time_step_len_seconds: the length of a time-step (period for logging)
        :param version: the version
        :param snort_ids_manager_port: the GRPC port of the snort IDS manager
        :param snort_ids_manager_log_file: the log file of the snort IDS manager
        :param snort_ids_manager_log_dir: the log dir of the snort IDS manager
        """
        self.time_step_len_seconds = time_step_len_seconds
        self.version = version
        self.snort_ids_manager_port = snort_ids_manager_port
        self.snort_ids_manager_log_file = snort_ids_manager_log_file
        self.snort_ids_manager_log_dir = snort_ids_manager_log_dir
        self.snort_ids_manager_max_workers = snort_ids_manager_max_workers

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SnortIDSManagerConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = SnortIDSManagerConfig(time_step_len_seconds=d["time_step_len_seconds"],
                                    version=d["version"], snort_ids_manager_port=d["snort_ids_manager_port"],
                                    snort_ids_manager_max_workers=d["snort_ids_manager_max_workers"],
                                    snort_ids_manager_log_dir=d["snort_ids_manager_log_dir"],
                                    snort_ids_manager_log_file=d["snort_ids_manager_log_file"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the object
        """
        d = {}
        d["snort_ids_manager_port"] = self.snort_ids_manager_port
        d["time_step_len_seconds"] = self.time_step_len_seconds
        d["version"] = self.version
        d["snort_ids_manager_max_workers"] = self.snort_ids_manager_max_workers
        d["snort_ids_manager_log_dir"] = self.snort_ids_manager_log_dir
        d["snort_ids_manager_log_file"] = self.snort_ids_manager_log_file
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"snort_ids_manager_port: {self.snort_ids_manager_port}, " \
               f"time_step_len_seconds: {self.time_step_len_seconds}," \
               f" version: {self.version}, snort_ids_manager_max_workers: {self.snort_ids_manager_max_workers}," \
               f"snort_ids_manager_log_file: {self.snort_ids_manager_log_file}, " \
               f"snort_ids_manager_log_dir: {self.snort_ids_manager_log_dir}"

    def to_json_str(self) -> str:
        """
        Converts the DTO into a json string

        :return: the json string representation of the DTO
        """
        import json
        json_str = json.dumps(self.to_dict(), indent=4, sort_keys=True)
        return json_str

    def to_json_file(self, json_file_path: str) -> None:
        """
        Saves the DTO to a json file

        :param json_file_path: the json file path to save  the DTO to
        :return: None
        """
        import io
        json_str = self.to_json_str()
        with io.open(json_file_path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    @staticmethod
    def from_json_file(json_file_path: str) -> "SnortIDSManagerConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return SnortIDSManagerConfig.from_dict(json.loads(json_str))

    def copy(self) -> "SnortIDSManagerConfig":
        """
        :return: a copy of the DTO
        """
        return SnortIDSManagerConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "SnortIDSManagerConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        return config

    @staticmethod
    def schema() -> "SnortIDSManagerConfig":
        """
        :return: get the schema of the DTO
        """
        return SnortIDSManagerConfig(snort_ids_manager_log_file="snort_ids_manager.log", snort_ids_manager_log_dir="/",
                                     snort_ids_manager_max_workers=10)



from typing import List
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig


class RandomizationSpaceConfig:
    """
    DTO representing a randomization space configuratio
    """

    def __init__(self, emulation_confs: List[EmulationEnvConfig], min_num_nodes: int = 4,
                 max_num_nodes: int = 4, min_num_flags: int = 1, max_num_flags: int = 1,
                 min_num_users: int = 1, max_num_users: int = 1,
                 services=None, vulnerabilities=None, os: str = None,
                 use_base_randomization: bool = False) -> None:
        """
        Creates a randomization space config according to the given parameters

        :param network_confs: the network configurations of the environment
        :param min_num_nodes: the minimum number of nodes in a sampled MDP
        :param max_num_nodes: the maximum number of nodes in a sampled MDP
        :param min_num_flags: the minimum number of flags in a sampled MDP
        :param max_num_flags: the maximum number of flags in a sampled MDP
        :param min_num_users: the minimum number of users in a sampled MDP
        :param max_num_users: the maximum number of users in a sampled MDP
        :param services: the list of possible services to include in the sampled MDP
        :param vulnerabilities: the list of possible vulnerabilities to include in the sampled MDP
        :param os: the list of possible operating systems to include in the sampled MDP
        :param use_base_randomization: boolean flag whether to use a base set of services/vulnerabilities
        :return: the created randomization space
        """
        self.emulation_confs = emulation_confs
        self.min_num_nodes = min_num_nodes
        self.max_num_nodes = max_num_nodes
        self.min_num_flags = min_num_flags
        self.max_num_flags = max_num_flags
        self.min_num_users = min_num_users
        self.services = services
        self.max_num_users = max_num_users
        self.vulnerabilities = vulnerabilities
        self.os = os
        self.use_base_randomization = use_base_randomization

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"min_num_nodes: {self.min_num_nodes}, max_num_nodes: {self.max_num_nodes}, " \
               f"min_num_flags:{self.min_num_flags}, max_num_flags: {self.max_num_flags}, " \
               f"min_num_users: {self.min_num_users}, max_num_users: {self.max_num_users}, " \
               f"services:{list(map(lambda x: str(x), self.services))}," \
               f"vulnerabilities: {list(map(lambda x: str(x), self.vulnerabilities))}, " \
               f"use_base_randomization: {self.use_base_randomization}," \
               f"network_confs:{list(map(lambda x: str(x), self.emulation_confs))}," \
               f"os:{list(map(lambda x: str(x), self.os))}"



"""
Type of process that generates new clients for an emulation
"""

from enum import IntEnum


class ClientPopulationProcessType(IntEnum):
    """
    Enum representing the different process types for generating client process in an emulation
    """
    POISSON = 0
    SINE_MODULATED_POISSON = 1



from enum import IntEnum


class AgentType(IntEnum):
    """
    Enum representing the different agent types in CSLE
    """
    T_SPSA = 0
    PPO = 1
    T_FP = 2
    DQN = 3
    REINFORCE = 4
    NFSP = 5
    RANDOM = 6
    NONE = 7
    VALUE_ITERATION = 8
    HSVI = 9
    SONDIK_VALUE_ITERATION = 10
    RANDOM_SEARCH = 11
    DIFFERENTIAL_EVOLUTION = 12
    CROSS_ENTROPY = 13
    KIEFER_WOLFOWITZ = 14
    Q_LEARNING = 15
    SARSA = 16
    POLICY_ITERATION = 17
    SHAPLEY_ITERATION = 18
    HSVI_OS_POSG = 19
    FICTITIOUS_PLAY = 20
    LINEAR_PROGRAMMING_NORMAL_FORM = 21
    DYNA_SEC = 22
    BAYESIAN_OPTIMIZATION = 23



from typing import Dict, Any, Union
from csle_common.dao.emulation_config.cluster_config import ClusterConfig
import csle_common.constants.constants as constants
import os
import io
import json


class Config:
    """
    A DTO Class representing the CSLE configuration
    """

    def __init__(self, management_admin_username_default: str, management_admin_password_default: str,
                 management_admin_first_name_default: str, management_admin_last_name_default: str,
                 management_admin_email_default: str, management_admin_organization_default: str,
                 management_guest_username_default: str, management_guest_password_default: str,
                 management_guest_first_name_default: str, management_guest_last_name_default: str,
                 management_guest_email_default: str, management_guest_organization_default: str,
                 ssh_admin_username: str, ssh_admin_password: str, ssh_agent_username: str, ssh_agent_password: str,
                 metastore_user: str, metastore_password: str, metastore_database_name: str, metastore_ip: str,
                 node_exporter_port: int, grafana_port: int,
                 management_system_port: int, cadvisor_port: int, prometheus_port: int, node_exporter_pid_file: str,
                 pgadmin_port: int,
                 csle_mgmt_webapp_pid_file: str, docker_stats_manager_log_file: str, docker_stats_manager_log_dir: str,
                 docker_stats_manager_port: int, docker_stats_manager_max_workers: int,
                 docker_stats_manager_outfile: str, docker_stats_manager_pidfile: str, prometheus_pid_file: str,
                 prometheus_log_file: str, prometheus_config_file: str, default_log_dir: str,
                 cluster_config: ClusterConfig, node_exporter_log_file: str,
                 allow_registration: bool, grafana_username: str, grafana_password: str,
                 pgadmin_username: str, pgadmin_password: str, postgresql_log_dir: str, nginx_log_dir: str,
                 flask_log_file: str, cluster_manager_log_file: str):
        """
        Initializes the DTO

        :param management_admin_username_default: the default admin username to the management system
        :param management_admin_password_default: the default admin password to the management system
        :param management_admin_first_name_default: the default first name of the admin in the management system
        :param management_admin_last_name_default: the default last name of the admin in the management system
        :param management_admin_email_default: the default email of the admin in the management system
        :param management_admin_organization_default: the default organization of the admin in the management system
        :param management_guest_username_default: the default username of the guest user in the management system
        :param management_guest_password_default: teh default password of the guest user in the management system
        :param management_guest_first_name_default: teh default first name of the guest user in the management system
        :param management_guest_last_name_default: the default last name of the guest user in the management system
        :param management_guest_email_default: the default email of the guest user in the management system
        :param management_guest_organization_default: the default organization of the guest user
                                                      in the management system
        :param ssh_admin_username: the default SSH admin username to containers in emulation environments
        :param ssh_admin_password: the default SSH admin password to containers in emulation environments
        :param ssh_agent_username: the default SSH agent username to containers in emulation environments
        :param ssh_agent_password: the default SSH agent password to containers in emulation environments
        :param metastore_user: the default user to the metastore
        :param metastore_password: the default password to the metastore
        :param metastore_database_name: the default metastore database name
        :param metastore_ip: the default ip to the metastore
        :param node_exporter_port: the default port to run node_exporter
        :param grafana_port: the default port to run grafana
        :param management_system_port: the default port to run the management system
        :param cadvisor_port: the default port to run c_advisor
        :param prometheus_port: the default port to run prometheus
        :param pgadmin_port: the default port to run pgadmin
        :param node_exporter_pid_file: the default port to run node_exporter
        :param csle_mgmt_webapp_pid_file: the file to save the PID of the mgmt webapp
        :param docker_stats_manager_log_file: the file to save the logs of docker statsmanager
        :param docker_stats_manager_log_dir: the directory to save the logs of the docker statsmanager
        :param docker_stats_manager_port: the port of the docker statsmanager
        :param docker_stats_manager_max_workers: the maximum number of GRPC workers of the docker statsmanager
        :param docker_stats_manager_outfile: the outputfile of the docker statsmanaer
        :param docker_stats_manager_pidfile: the file to save the PID of the docker statsmanager
        :param prometheus_pid_file: the file to save the PID of prometheus
        :param prometheus_log_file: the log file of prometheus
        :param prometheus_config_file: the the config file of prometheus
        :param default_log_dir: the default log directory for CSLE applications
        :param cluster_config: the cluster configuration of the CSLE deployment
        :param node_exporter_log_file: the file to save the logs of the node_exporter
        :param allow_registration: boolean flag indicating whether user registration should be allowed
        :param grafana_username: default grafana username
        :param grafana_password: default grafana password
        :param pgadmin_username: default pgadmin username
        :param pgadmin_password: default pgadmin password
        :param postgresql_log_dir: the log directory for postgresql
        :param nginx_log_dir: the log directory for Nginx
        :param flask_log_file: the log file of Flask
        :param cluster_manager_log_file: the log file of the cluster manager
        """
        self.management_admin_username_default = management_admin_username_default
        self.management_admin_password_default = management_admin_password_default
        self.management_admin_first_name_default = management_admin_first_name_default
        self.management_admin_last_name_default = management_admin_last_name_default
        self.management_admin_email_default = management_admin_email_default
        self.management_admin_organization_default = management_admin_organization_default
        self.management_guest_username_default = management_guest_username_default
        self.management_guest_password_default = management_guest_password_default
        self.management_guest_first_name_default = management_guest_first_name_default
        self.management_guest_last_name_default = management_guest_last_name_default
        self.management_guest_email_default = management_guest_email_default
        self.management_guest_organization_default = management_guest_organization_default
        self.ssh_admin_username = ssh_admin_username
        self.ssh_admin_password = ssh_admin_password
        self.ssh_agent_username = ssh_agent_username
        self.ssh_agent_password = ssh_agent_password
        self.metastore_user = metastore_user
        self.metastore_password = metastore_password
        self.metastore_database_name = metastore_database_name
        self.metastore_ip = metastore_ip
        self.node_exporter_port = node_exporter_port
        self.grafana_port = grafana_port
        self.management_system_port = management_system_port
        self.cadvisor_port = cadvisor_port
        self.prometheus_port = prometheus_port
        self.pgadmin_port = pgadmin_port
        self.node_exporter_pid_file = node_exporter_pid_file
        self.csle_mgmt_webapp_pid_file = csle_mgmt_webapp_pid_file
        self.docker_stats_manager_log_file = docker_stats_manager_log_file
        self.docker_stats_manager_log_dir = docker_stats_manager_log_dir
        self.docker_stats_manager_port = docker_stats_manager_port
        self.docker_stats_manager_max_workers = docker_stats_manager_max_workers
        self.docker_stats_manager_outfile = docker_stats_manager_outfile
        self.docker_stats_manager_pidfile = docker_stats_manager_pidfile
        self.prometheus_pid_file = prometheus_pid_file
        self.prometheus_log_file = prometheus_log_file
        self.prometheus_config_file = prometheus_config_file
        self.default_log_dir = default_log_dir
        self.cluster_config = cluster_config
        self.node_exporter_log_file = node_exporter_log_file
        self.allow_registration = allow_registration
        self.grafana_username = grafana_username
        self.grafana_password = grafana_password
        self.pgadmin_username = pgadmin_username
        self.pgadmin_password = pgadmin_password
        self.postgresql_log_dir = postgresql_log_dir
        self.nginx_log_dir = nginx_log_dir
        self.flask_log_file = flask_log_file
        self.cluster_manager_log_file = cluster_manager_log_file
        self.id = 1

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the object
        """
        d = {}
        d["management_admin_username_default"] = self.management_admin_username_default
        d["management_admin_password_default"] = self.management_admin_password_default
        d["management_admin_first_name_default"] = self.management_admin_first_name_default
        d["management_admin_last_name_default"] = self.management_admin_last_name_default
        d["management_admin_email_default"] = self.management_admin_email_default
        d["management_admin_organization_default"] = self.management_admin_organization_default
        d["management_guest_username_default"] = self.management_guest_username_default
        d["management_guest_password_default"] = self.management_guest_password_default
        d["management_guest_first_name_default"] = self.management_guest_first_name_default
        d["management_guest_last_name_default"] = self.management_guest_last_name_default
        d["management_guest_email_default"] = self.management_guest_email_default
        d["management_guest_organization_default"] = self.management_guest_organization_default
        d["ssh_admin_username"] = self.ssh_admin_username
        d["ssh_admin_password"] = self.ssh_admin_password
        d["ssh_agent_username"] = self.ssh_agent_username
        d["ssh_agent_password"] = self.ssh_agent_password
        d["metastore_user"] = self.metastore_user
        d["metastore_password"] = self.metastore_password
        d["metastore_database_name"] = self.metastore_database_name
        d["metastore_ip"] = self.metastore_ip
        d["node_exporter_port"] = self.node_exporter_port
        d["grafana_port"] = self.grafana_port
        d["management_system_port"] = self.management_system_port
        d["cadvisor_port"] = self.cadvisor_port
        d["pgadmin_port"] = self.pgadmin_port
        d["prometheus_port"] = self.prometheus_port
        d["node_exporter_pid_file"] = self.node_exporter_pid_file
        d["csle_mgmt_webapp_pid_file"] = self.csle_mgmt_webapp_pid_file
        d["docker_stats_manager_log_file"] = self.docker_stats_manager_log_file
        d["docker_stats_manager_log_dir"] = self.docker_stats_manager_log_dir
        d["docker_stats_manager_port"] = self.docker_stats_manager_port
        d["docker_stats_manager_max_workers"] = self.docker_stats_manager_max_workers
        d["docker_stats_manager_outfile"] = self.docker_stats_manager_outfile
        d["docker_stats_manager_pidfile"] = self.docker_stats_manager_pidfile
        d["prometheus_pid_file"] = self.prometheus_pid_file
        d["prometheus_log_file"] = self.prometheus_log_file
        d["prometheus_config_file"] = self.prometheus_config_file
        d["default_log_dir"] = self.default_log_dir
        d["cluster_config"] = self.cluster_config.to_dict()
        d["node_exporter_log_file"] = self.node_exporter_log_file
        d["allow_registration"] = self.allow_registration
        d["grafana_username"] = self.grafana_username
        d["grafana_password"] = self.grafana_password
        d["pgadmin_username"] = self.pgadmin_username
        d["pgadmin_password"] = self.pgadmin_password
        d["postgresql_log_dir"] = self.postgresql_log_dir
        d["nginx_log_dir"] = self.nginx_log_dir
        d["flask_log_file"] = self.flask_log_file
        d["cluster_manager_log_file"] = self.cluster_manager_log_file
        d["id"] = 1
        return d

    def to_param_dict(self) -> Dict[str, Any]:
        """
        :return: a param-dict representation of the object
        """
        d = {}
        d["parameters"] = []
        d["parameters"].append(
            {
                "id": 0,
                "param": "management_admin_username_default",
                "value": self.management_admin_username_default
            }
        )
        d["parameters"].append(
            {
                "id": 1,
                "param": "management_admin_password_default",
                "value": self.management_admin_password_default
            }
        )
        d["parameters"].append(
            {
                "id": 2,
                "param": "management_admin_first_name_default",
                "value": self.management_admin_first_name_default
            }
        )
        d["parameters"].append(
            {
                "id": 3,
                "param": "management_admin_last_name_default",
                "value": self.management_admin_last_name_default
            }
        )
        d["parameters"].append(
            {
                "id": 4,
                "param": "management_admin_email_default",
                "value": self.management_admin_email_default
            }
        )
        d["parameters"].append(
            {
                "id": 5,
                "param": "management_admin_organization_default",
                "value": self.management_admin_organization_default
            }
        )
        d["parameters"].append(
            {
                "id": 6,
                "param": "management_guest_username_default",
                "value": self.management_guest_username_default
            }
        )
        d["parameters"].append(
            {
                "id": 7,
                "param": "management_guest_password_default",
                "value": self.management_guest_password_default
            }
        )
        d["parameters"].append(
            {
                "id": 8,
                "param": "management_guest_first_name_default",
                "value": self.management_guest_first_name_default
            }
        )
        d["parameters"].append(
            {
                "id": 9,
                "param": "management_guest_last_name_default",
                "value": self.management_guest_last_name_default
            }
        )
        d["parameters"].append(
            {
                "id": 10,
                "param": "management_guest_email_default",
                "value": self.management_guest_email_default
            }
        )
        d["parameters"].append(
            {
                "id": 11,
                "param": "management_guest_organization_default",
                "value": self.management_guest_organization_default
            }
        )
        d["parameters"].append(
            {
                "id": 12,
                "param": "ssh_admin_username",
                "value": self.ssh_admin_username
            }
        )
        d["parameters"].append(
            {
                "id": 13,
                "param": "ssh_admin_password",
                "value": self.ssh_admin_password
            }
        )
        d["parameters"].append(
            {
                "id": 14,
                "param": "ssh_agent_username",
                "value": self.ssh_agent_username
            }
        )
        d["parameters"].append(
            {
                "id": 15,
                "param": "ssh_agent_password",
                "value": self.ssh_agent_password
            }
        )
        d["parameters"].append(
            {
                "id": 16,
                "param": "metastore_user",
                "value": self.metastore_user
            }
        )
        d["parameters"].append(
            {
                "id": 17,
                "param": "metastore_password",
                "value": self.metastore_password
            }
        )
        d["parameters"].append(
            {
                "id": 18,
                "param": "metastore_database_name",
                "value": self.metastore_database_name
            }
        )
        d["parameters"].append(
            {
                "id": 19,
                "param": "metastore_ip",
                "value": self.metastore_ip
            }
        )
        d["parameters"].append(
            {
                "id": 20,
                "param": "node_exporter_port",
                "value": self.node_exporter_port
            }
        )
        d["parameters"].append(
            {
                "id": 21,
                "param": "grafana_port",
                "value": self.grafana_port
            }
        )
        d["parameters"].append(
            {
                "id": 23,
                "param": "management_system_port",
                "value": self.management_system_port
            }
        )
        d["parameters"].append(
            {
                "id": 24,
                "param": "cadvisor_port",
                "value": self.cadvisor_port
            }
        )
        d["parameters"].append(
            {
                "id": 25,
                "param": "prometheus_port",
                "value": self.prometheus_port
            }
        )
        d["parameters"].append(
            {
                "id": 26,
                "param": "node_exporter_pid_file",
                "value": self.node_exporter_pid_file
            }
        )
        d["parameters"].append(
            {
                "id": 27,
                "param": "csle_mgmt_webapp_pid_file",
                "value": self.csle_mgmt_webapp_pid_file
            }
        )
        d["parameters"].append(
            {
                "id": 28,
                "param": "docker_stats_manager_log_file",
                "value": self.docker_stats_manager_log_file
            }
        )
        d["parameters"].append(
            {
                "id": 29,
                "param": "docker_stats_manager_log_dir",
                "value": self.docker_stats_manager_log_dir
            }
        )
        d["parameters"].append(
            {
                "id": 30,
                "param": "docker_stats_manager_port",
                "value": self.docker_stats_manager_port
            }
        )
        d["parameters"].append(
            {
                "id": 31,
                "param": "docker_stats_manager_max_workers",
                "value": self.docker_stats_manager_max_workers
            }
        )
        d["parameters"].append(
            {
                "id": 32,
                "param": "docker_stats_manager_outfile",
                "value": self.docker_stats_manager_outfile
            }
        )
        d["parameters"].append(
            {
                "id": 33,
                "param": "docker_stats_manager_pidfile",
                "value": self.docker_stats_manager_pidfile
            }
        )
        d["parameters"].append(
            {
                "id": 34,
                "param": "prometheus_pid_file",
                "value": self.prometheus_pid_file
            }
        )
        d["parameters"].append(
            {
                "id": 35,
                "param": "prometheus_log_file",
                "value": self.prometheus_log_file
            }
        )
        d["parameters"].append(
            {
                "id": 36,
                "param": "prometheus_config_file",
                "value": self.prometheus_config_file
            }
        )
        d["parameters"].append(
            {
                "id": 37,
                "param": "default_log_dir",
                "value": self.default_log_dir
            }
        )
        d["parameters"].append(
            {
                "id": 39,
                "param": "node_exporter_log_file",
                "value": self.node_exporter_log_file
            }
        ),
        d["parameters"].append(
            {
                "id": 40,
                "param": "allow_registration",
                "value": self.allow_registration
            }
        )
        d["parameters"].append(
            {
                "id": 41,
                "param": "pgadmin_port",
                "value": self.pgadmin_port
            }
        )
        d["parameters"].append(
            {
                "id": 42,
                "param": "pgadmin_username",
                "value": self.pgadmin_username
            }
        )
        d["parameters"].append(
            {
                "id": 43,
                "param": "pgadmin_password",
                "value": self.pgadmin_password
            }
        )
        d["parameters"].append(
            {
                "id": 44,
                "param": "grafana_username",
                "value": self.grafana_username
            }
        )
        d["parameters"].append(
            {
                "id": 45,
                "param": "grafana_password",
                "value": self.grafana_password
            }
        )
        d["parameters"].append(
            {
                "id": 46,
                "param": "postgresql_log_dir",
                "value": self.postgresql_log_dir
            }
        )
        d["parameters"].append(
            {
                "id": 47,
                "param": "nginx_log_dir",
                "value": self.nginx_log_dir
            }
        )
        d["parameters"].append(
            {
                "id": 48,
                "param": "flask_log_file",
                "value": self.flask_log_file
            }
        )
        d["parameters"].append(
            {
                "id": 48,
                "param": "cluster_manager_log_file",
                "value": self.cluster_manager_log_file
            }
        )
        d["cluster_config"] = self.cluster_config.to_dict()
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "Config":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = Config(management_admin_username_default=d["management_admin_username_default"],
                     management_admin_password_default=d["management_admin_password_default"],
                     management_admin_first_name_default=d["management_admin_first_name_default"],
                     management_admin_last_name_default=d["management_admin_last_name_default"],
                     management_admin_email_default=d["management_admin_email_default"],
                     management_admin_organization_default=d["management_admin_organization_default"],
                     management_guest_username_default=d["management_guest_username_default"],
                     management_guest_password_default=d["management_guest_password_default"],
                     management_guest_first_name_default=d["management_guest_first_name_default"],
                     management_guest_last_name_default=d["management_guest_last_name_default"],
                     management_guest_email_default=d["management_guest_email_default"],
                     management_guest_organization_default=d["management_guest_organization_default"],
                     ssh_admin_username=d["ssh_admin_username"],
                     ssh_admin_password=d["ssh_admin_password"],
                     ssh_agent_username=d["ssh_agent_username"],
                     ssh_agent_password=d["ssh_agent_password"],
                     metastore_user=d["metastore_user"],
                     metastore_password=d["metastore_password"],
                     metastore_database_name=d["metastore_database_name"],
                     metastore_ip=d["metastore_ip"],
                     node_exporter_port=d["node_exporter_port"],
                     grafana_port=d["grafana_port"],
                     management_system_port=d["management_system_port"],
                     cadvisor_port=d["cadvisor_port"],
                     pgadmin_port=d["pgadmin_port"],
                     prometheus_port=d["prometheus_port"],
                     node_exporter_pid_file=d["node_exporter_pid_file"],
                     csle_mgmt_webapp_pid_file=d["csle_mgmt_webapp_pid_file"],
                     docker_stats_manager_log_file=d["docker_stats_manager_log_file"],
                     docker_stats_manager_log_dir=d["docker_stats_manager_log_dir"],
                     docker_stats_manager_port=d["docker_stats_manager_port"],
                     docker_stats_manager_max_workers=d["docker_stats_manager_max_workers"],
                     docker_stats_manager_outfile=d["docker_stats_manager_outfile"],
                     docker_stats_manager_pidfile=d["docker_stats_manager_pidfile"],
                     prometheus_pid_file=d["prometheus_pid_file"],
                     prometheus_log_file=d["prometheus_log_file"],
                     prometheus_config_file=d["prometheus_config_file"],
                     default_log_dir=d["default_log_dir"],
                     cluster_config=ClusterConfig.from_dict(d["cluster_config"]),
                     node_exporter_log_file=d["node_exporter_log_file"],
                     allow_registration=d["allow_registration"],
                     grafana_username=d["grafana_username"],
                     grafana_password=d["grafana_password"],
                     pgadmin_username=d["pgadmin_username"],
                     pgadmin_password=d["pgadmin_password"],
                     postgresql_log_dir=d["postgresql_log_dir"],
                     nginx_log_dir=d["nginx_log_dir"],
                     flask_log_file=d["flask_log_file"],
                     cluster_manager_log_file=d["cluster_manager_log_file"]
                     )
        if "id" in d:
            dto.id = d["id"]
        return dto

    @staticmethod
    def from_param_dict(d2: Dict[str, Any]) -> "Config":
        """
        Convert a param-dict representation to a DTO representation

        :return: a dto representation of the object
        """
        d = {}
        d["cluster_config"] = d2["cluster_config"]
        for param_value in d2["parameters"]:
            d[param_value["param"]] = param_value["value"]

        dto = Config(management_admin_username_default=d["management_admin_username_default"],
                     management_admin_password_default=d["management_admin_password_default"],
                     management_admin_first_name_default=d["management_admin_first_name_default"],
                     management_admin_last_name_default=d["management_admin_last_name_default"],
                     management_admin_email_default=d["management_admin_email_default"],
                     management_admin_organization_default=d["management_admin_organization_default"],
                     management_guest_username_default=d["management_guest_username_default"],
                     management_guest_password_default=d["management_guest_password_default"],
                     management_guest_first_name_default=d["management_guest_first_name_default"],
                     management_guest_last_name_default=d["management_guest_last_name_default"],
                     management_guest_email_default=d["management_guest_email_default"],
                     management_guest_organization_default=d["management_guest_organization_default"],
                     ssh_admin_username=d["ssh_admin_username"],
                     ssh_admin_password=d["ssh_admin_password"],
                     ssh_agent_username=d["ssh_agent_username"],
                     ssh_agent_password=d["ssh_agent_password"],
                     metastore_user=d["metastore_user"],
                     metastore_password=d["metastore_password"],
                     metastore_database_name=d["metastore_database_name"],
                     metastore_ip=d["metastore_ip"],
                     node_exporter_port=d["node_exporter_port"],
                     grafana_port=d["grafana_port"],
                     management_system_port=d["management_system_port"],
                     cadvisor_port=d["cadvisor_port"],
                     prometheus_port=d["prometheus_port"],
                     node_exporter_pid_file=d["node_exporter_pid_file"],
                     csle_mgmt_webapp_pid_file=d["csle_mgmt_webapp_pid_file"],
                     docker_stats_manager_log_file=d["docker_stats_manager_log_file"],
                     docker_stats_manager_log_dir=d["docker_stats_manager_log_dir"],
                     docker_stats_manager_port=d["docker_stats_manager_port"],
                     docker_stats_manager_max_workers=d["docker_stats_manager_max_workers"],
                     docker_stats_manager_outfile=d["docker_stats_manager_outfile"],
                     docker_stats_manager_pidfile=d["docker_stats_manager_pidfile"],
                     prometheus_pid_file=d["prometheus_pid_file"],
                     prometheus_log_file=d["prometheus_log_file"],
                     prometheus_config_file=d["prometheus_config_file"],
                     default_log_dir=d["default_log_dir"],
                     cluster_config=ClusterConfig.from_dict(d["cluster_config"]),
                     node_exporter_log_file=d["node_exporter_log_file"],
                     allow_registration=d["allow_registration"],
                     pgadmin_port=d["pgadmin_port"],
                     grafana_username=d["grafana_username"],
                     grafana_password=d["grafana_password"],
                     pgadmin_username=d["pgadmin_username"],
                     pgadmin_password=d["pgadmin_password"],
                     postgresql_log_dir=d["postgresql_log_dir"],
                     nginx_log_dir=d["nginx_log_dir"],
                     flask_log_file=d["flask_log_file"],
                     cluster_manager_log_file=d["cluster_manager_log_file"])
        if "id" in d:
            dto.id = d["id"]
        return dto

    def __str__(self) -> str:
        """
        :return: a string representation of the credential
        """
        return f"management_admin_username_default:{self.management_admin_username_default}, " \
               f"management_admin_password_default: {self.management_admin_password_default}," \
               f"management_admin_first_name_default: {self.management_admin_first_name_default}, " \
               f"management_admin_last_name_default: {self.management_admin_last_name_default}," \
               f"management_admin_email_default: {self.management_admin_email_default}, " \
               f"management_admin_organization_default: {self.management_admin_organization_default}, " \
               f"ssh_admin_username: {self.ssh_admin_username}, " \
               f"ssh_admin_password: {self.ssh_admin_password}, " \
               f"ssh_agent_username: {self.ssh_agent_username}, " \
               f"ssh_agent_password: {self.ssh_agent_password}, " \
               f"metastore_user: {self.metastore_user}, " \
               f"metastore_password: {self.metastore_password}, " \
               f"metastore_database_name: {self.metastore_database_name}, " \
               f"metastore_ip: {self.metastore_ip}, " \
               f"node_exporter_port: {self.node_exporter_port}, " \
               f"grafana_port: {self.grafana_port}, " \
               f"management_system_port: {self.management_system_port}, " \
               f"cadvisor_port: {self.cadvisor_port}, " \
               f"prometheus_port: {self.prometheus_port}, " \
               f"node_exporter_pid_file: {self.node_exporter_pid_file}, " \
               f"csle_mgmt_webapp_pid_file: {self.csle_mgmt_webapp_pid_file}, " \
               f"docker_stats_manager_log_file: {self.docker_stats_manager_log_file}, " \
               f"docker_stats_manager_log_dir: {self.docker_stats_manager_log_dir}, " \
               f"docker_stats_manager_port: {self.docker_stats_manager_port}, " \
               f"docker_stats_manager_max_workers: {self.docker_stats_manager_max_workers}, " \
               f"docker_stats_manager_outfile: {self.docker_stats_manager_outfile}, " \
               f"docker_stats_manager_pidfile: {self.docker_stats_manager_pidfile}, " \
               f"prometheus_pid_file: {self.prometheus_pid_file}, " \
               f"prometheus_log_file: {self.prometheus_log_file}, " \
               f"prometheus_config_file: {self.prometheus_config_file}, " \
               f"default_log_dir: {self.default_log_dir}, " \
               f"cluster_config: {self.cluster_config}," \
               f"node_exporter_log_file: {self.node_exporter_log_file}," \
               f"allow_registration: {self.allow_registration}, " \
               f"id:{self.id}, grafana_username: {self.grafana_username}, grafana_password: {self.grafana_password}," \
               f"pgadmin_username: {self.pgadmin_username}, pgadmin_password: {self.pgadmin_password}," \
               f"postgresql_log_dir: {self.postgresql_log_dir}, nginx_log_dir: {self.nginx_log_dir}," \
               f"flask_log_file: {self.flask_log_file}, cluster_manager_log_file: {self.cluster_manager_log_file}"

    def to_json_str(self) -> str:
        """
        Converts the DTO into a json string

        :return: the json string representation of the DTO
        """
        import json
        json_str = json.dumps(self.to_dict(), indent=4, sort_keys=True)
        return json_str

    def to_json_file(self, json_file_path: str) -> None:
        """
        Saves the DTO to a json file

        :param json_file_path: the json file path to save  the DTO to
        :return: None
        """
        import io
        json_str = self.to_json_str()
        with io.open(json_file_path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    @staticmethod
    def from_json_file(json_file_path: str) -> "Config":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return Config.from_dict(json.loads(json_str))

    def copy(self) -> "ClusterConfig":
        """
        :return: a copy of the DTO
        """
        return ClusterConfig.from_dict(self.to_dict())

    @staticmethod
    def read_config_file() -> "Config":
        """
        Reads and parses the config file

        :return: the parsed config file
        """
        if constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM in os.environ:
            csle_home = os.environ[constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM]
        else:
            raise Exception(f"The environment parameter {constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM} is not set")
        config_file_path = f"{csle_home}{constants.COMMANDS.SLASH_DELIM}" \
                           f"{constants.CONFIG_FILE.CONFIG_FILE_NAME}"
        with io.open(config_file_path, 'r', encoding='utf-8') as f:
            json_str = f.read()
            config_dict = json.loads(json_str)
            config = Config.from_dict(config_dict)
            return config

    @staticmethod
    def save_config_file(config: Dict[str, Any]) -> None:
        """
        Saves the config file to disk

        :return: None
        """
        if constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM in os.environ:
            csle_home = os.environ[constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM]
        else:
            raise Exception(f"The environment parameter {constants.CONFIG_FILE.CSLE_HOME_ENV_PARAM} is not set")
        config_file_path = f"{csle_home}{constants.COMMANDS.SLASH_DELIM}" \
                           f"{constants.CONFIG_FILE.CONFIG_FILE_NAME}"
        with io.open(config_file_path, 'w', encoding='utf-8') as f:
            json_str = json.dumps(config, indent=4, sort_keys=True)
            f.write(json_str)

    @staticmethod
    def get_current_config() -> Union["Config", None]:
        """
        :return: The currently parsed config
        """
        return constants.CONFIG_FILE.PARSED_CONFIG



from typing import List, Tuple, Set


class EmulationEnvGenerationConfig:
    """
    A DTO representing the configuration of a dynamically created emulation environment
    """

    def __init__(self, min_num_users: int = 1, max_num_users: int = 5, min_num_flags: int = 1,
                 max_num_flags: int = 5, min_num_nodes: int = 4, max_num_nodes: int = 10,
                 container_pool: List[Tuple[str, str]] = None,
                 gw_vuln_compatible_containers: List[Tuple[str, str]] = None,
                 pw_vuln_compatible_containers: List[Tuple[str, str]] = None,
                 rce_vuln_compatible_containers: List[Tuple[str, str]] = None,
                 sql_injection_vuln_compatible_containers: List[Tuple[str, str]] = None,
                 priv_esc_vuln_compatible_containers: List[Tuple[str, str]] = None,
                 agent_containers: List[Tuple[str, str]] = None,
                 router_containers: List[Tuple[str, str]] = None,
                 path: str = "", subnet_id_blacklist: Set = None,
                 subnet_prefix: str = "", subnet_id: int = -1, num_flags: int = -1, num_nodes: int = -1,
                 min_cpus: int = 1, max_cpus: int = 1, min_mem_G: int = 4, max_mem_G: int = 4):
        """
        Initializes the object

        :param min_num_users: the minimum number of users
        :param max_num_users: the maximum number of users
        :param min_num_flags: the minimum number of flags
        :param max_num_flags: the maximum number of flags
        :param min_num_nodes: the minimum number of nodes
        :param max_num_nodes: the maximum number of nodes
        :param container_pool: the pool of containers for emulation
        :param gw_vuln_compatible_containers: the pool of gw-compatible containers
        :param pw_vuln_compatible_containers: the pool of pw-vuln-compatible containers
        :param rce_vuln_compatible_containers: the pool of rce-compatible containers
        :param sql_injection_vuln_compatible_containers: the pool of sql-injection-compatible containers
        :param priv_esc_vuln_compatible_containers: the pool of priv-esc-compatible containers
        :param agent_containers: the pool of agent containers
        :param router_containers: the pool of router containers
        :param path: the path to the config
        :param subnet_id_blacklist: the black list of subnet ids
        :param subnet_prefix: the prefix of the subnet
        :param subnet_id: the subnet id
        :param num_flags: the number of flags
        :param num_nodes: the number of nodes
        :param min_cpus: the minimum number of CPUs per container
        :param max_cpus: the maximum number of CPUs per container
        :param min_mem_G: the minimum number of memory GB for each container
        :param max_mem_G: the maximum number of memory GB for each container
        """
        self.min_num_users = min_num_users
        self.max_num_users = max_num_users
        self.min_num_flags = min_num_flags
        self.max_num_flags = max_num_flags
        self.min_num_nodes = min_num_nodes
        self.max_num_nodes = max_num_nodes
        self.container_pool = container_pool
        self.gw_vuln_compatible_containers = gw_vuln_compatible_containers
        self.pw_vuln_compatible_containers = pw_vuln_compatible_containers
        self.rce_vuln_compatible_containers = rce_vuln_compatible_containers
        self.sql_injection_vuln_compatible_containers = sql_injection_vuln_compatible_containers
        self.priv_esc_vuln_compatible_containers = priv_esc_vuln_compatible_containers
        self.agent_containers = agent_containers
        self.router_containers = router_containers
        self.path = path
        self.subnet_id_blacklist = subnet_id_blacklist
        self.subnet_prefix = subnet_prefix
        self.subnet_id = subnet_id
        self.num_flags = num_flags
        self.num_nodes = num_nodes
        self.min_cpus = min_cpus
        self.max_cpus = max_cpus
        self.min_mem_G = min_mem_G
        self.max_mem_G = max_mem_G

        if self.container_pool is None:
            self.container_pool = []
        if self.gw_vuln_compatible_containers is None:
            self.gw_vuln_compatible_containers = []
        if self.pw_vuln_compatible_containers is None:
            self.pw_vuln_compatible_containers = []
        if self.rce_vuln_compatible_containers is None:
            self.rce_vuln_compatible_containers = []
        if self.sql_injection_vuln_compatible_containers is None:
            self.sql_injection_vuln_compatible_containers = []
        if self.priv_esc_vuln_compatible_containers is None:
            self.priv_esc_vuln_compatible_containers = []
        if self.agent_containers is None:
            self.agent_containers = []
        if self.router_containers is None:
            self.router_containers = []
        if self.subnet_id_blacklist is None:
            self.subnet_id_blacklist = set()

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"min_num_users:{self.min_num_users}, max_num_users:{self.max_num_users}, " \
               f"min_num_flags:{self.min_num_flags}, max_num_flags:{self.max_num_flags}," \
               f"min_num_nodes:{self.min_num_nodes}, max_num_nodes:{self.max_num_nodes}, " \
               f"container_pool:{self.container_pool}, " \
               f"gw_vuln_compatible_containers:{self.gw_vuln_compatible_containers}," \
               f"pw_vuln_compatible_containers:{self.pw_vuln_compatible_containers}, " \
               f"rce_vuln_compatible_containers:{self.rce_vuln_compatible_containers}, " \
               f"sql_injection_vuln_compatible_containers:{self.sql_injection_vuln_compatible_containers}, " \
               f"priv_esc_vuln_compatible_containers:{self.priv_esc_vuln_compatible_containers}" \
               f"agent_containers:{self.agent_containers}, router_containers:{self.router_containers}, " \
               f"path:{self.path}, subnet_id_blacklist:{self.subnet_id_blacklist}, " \
               f"subnet_prefix:{self.subnet_prefix}, subnet_id:{self.subnet_id}, num_flags:{self.num_flags}, " \
               f"num_nodes:{self.num_nodes}, min_cpus:{self.min_cpus}, max_cpus:{self.max_cpus}, " \
               f"min_mem_G:{self.min_mem_G}, max_mem_G:{self.max_mem_G}"

    def copy(self) -> "EmulationEnvGenerationConfig":
        """
        :return: a copy of the object
        """
        return EmulationEnvGenerationConfig(
            min_num_users=self.min_num_users, max_num_users=self.max_num_users, min_num_flags=self.min_num_flags,
            max_num_flags=self.max_num_flags, min_num_nodes=self.min_num_nodes, max_num_nodes=self.max_num_nodes,
            container_pool=self.container_pool, gw_vuln_compatible_containers=self.gw_vuln_compatible_containers,
            pw_vuln_compatible_containers=self.pw_vuln_compatible_containers,
            rce_vuln_compatible_containers=self.rce_vuln_compatible_containers,
            sql_injection_vuln_compatible_containers=self.sql_injection_vuln_compatible_containers,
            priv_esc_vuln_compatible_containers=self.priv_esc_vuln_compatible_containers,
            agent_containers=self.agent_containers, router_containers=self.router_containers,
            path=self.path, subnet_id_blacklist=self.subnet_id_blacklist.copy(),
            subnet_prefix=self.subnet_prefix, subnet_id=self.subnet_id, num_flags=self.num_flags,
            num_nodes=self.num_nodes, min_cpus=self.min_cpus, max_cpus=self.max_cpus,
            min_mem_G=self.min_mem_G, max_mem_G=self.max_mem_G
        )



from typing import Dict, Any
from csle_common.dao.system_identification.system_model import SystemModel
from csle_common.dao.system_identification.gaussian_mixture_system_model import GaussianMixtureSystemModel
from csle_common.dao.system_identification.system_identification_config import SystemIdentificationConfig


class SystemIdentificationJobConfig:
    """
    DTO representing a system identification job
    """

    def __init__(self, emulation_env_name: str, emulation_statistics_id: int,
                 progress_percentage: float, pid: int, log_file_path: str,
                 system_identification_config: SystemIdentificationConfig,
                 descr: str = "",
                 system_model: SystemModel = None):
        """
        Initializes the DTO

        :param emulation_env_name: the name of the emulation that the system identification concerns
        :param emulation_statistics_id: the id of the statistics data to train with
        :param progress_percentage: the progress percentage
        :param pid: the pid of the process
        :param log_file_path: path to the log file
        :param descr: a description of the job
        :param system_model: fitted system model
        :param system_identification_config: the config of the system identification algorithm
        """
        self.emulation_env_name = emulation_env_name
        self.emulation_statistics_id = emulation_statistics_id
        self.progress_percentage = progress_percentage
        self.pid = pid
        self.log_file_path = log_file_path
        self.descr = descr
        self.system_model = system_model
        self.system_identification_config = system_identification_config
        self.id = -1
        self.running = False

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the object
        """
        d = {}
        d["emulation_env_name"] = self.emulation_env_name
        d["pid"] = self.pid
        d["progress_percentage"] = self.progress_percentage
        d["emulation_statistics_id"] = self.emulation_statistics_id
        d["descr"] = self.descr
        d["log_file_path"] = self.log_file_path
        if self.system_model is None:
            d["system_model"] = None
        else:
            d["system_model"] = self.system_model.to_dict()
        d["system_identification_config"] = self.system_identification_config.to_dict()
        d["id"] = self.id
        d["running"] = self.running
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SystemIdentificationJobConfig":
        """
        Converts a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        system_model = None
        system_model = GaussianMixtureSystemModel.from_dict(d["system_model"])
        obj = SystemIdentificationJobConfig(
            emulation_env_name=d["emulation_env_name"], pid=d["pid"],
            progress_percentage=d["progress_percentage"], emulation_statistics_id=d["emulation_statistics_id"],
            descr=d["descr"], log_file_path=d["log_file_path"], system_model=system_model,
            system_identification_config=SystemIdentificationConfig.from_dict(d["system_identification_config"])
        )
        if "id" in d:
            obj.id = d["id"]
        if "running" in d:
            obj.running = d["running"]
        return obj

    def to_json_str(self) -> str:
        """
        Converts the DTO into a json string

        :return: the json string representation of the DTO
        """
        import json
        json_str = json.dumps(self.to_dict(), indent=4, sort_keys=True)
        return json_str

    def to_json_file(self, json_file_path: str) -> None:
        """
        Saves the DTO to a json file

        :param json_file_path: the json file path to save  the DTO to
        :return: None
        """
        import io
        json_str = self.to_json_str()
        with io.open(json_file_path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    @staticmethod
    def from_json_file(json_file_path: str) -> "SystemIdentificationJobConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return SystemIdentificationJobConfig.from_dict(json.loads(json_str))




class NmapHttpGrep:
    """
    A DTO representing the output of a Nmap HTTP Grep
    """

    def __init__(self, output: str):
        """
        Initializes the DTO

        :param output: the output of the HTTP Grep
        """
        self.output = output

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"output:{self.output}"



from csle_common.dao.domain_randomization.randomization_space_config import RandomizationSpaceConfig


class RandomizationSpace:
    """
    Object representing a randomization space
    """

    def __init__(self, config: RandomizationSpaceConfig):
        """
        Initializes the randomization space

        :param config: the randomization space config
        """
        self.config = config

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"config:{str(self.config)}"



from typing import List, Dict, Any


class ObservationFunctionConfig:
    """
    DTO representing the configuration of the observation function of a simulation
    """

    def __init__(self, observation_tensor: List, component_observation_tensors: Dict[str, List]):
        """
        Initializes the DTO
        :param observation_tensor: the observation tensor
        """
        self.observation_tensor = observation_tensor
        self.component_observation_tensors = component_observation_tensors

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ObservationFunctionConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ObservationFunctionConfig(
            observation_tensor=d["observation_tensor"],
            component_observation_tensors=d["component_observation_tensors"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the object
        """
        d = {}
        d["observation_tensor"] = self.observation_tensor
        d["component_observation_tensors"] = self.component_observation_tensors
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"observation tensor: {self.observation_tensor}, " \
               f"component_observation_tensors: {self.component_observation_tensors}"



from typing import Union
from csle_common.dao.emulation_observation.common.emulation_vulnerability_observation_state \
    import EmulationVulnerabilityObservationState
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol


class NiktoVuln:
    """
    DTO representing a vulnerability found with a Nikto scan
    """

    def __init__(self, id: str, osvdb_id: Union[int, None], method: str, iplink: str,
                 namelink: str, uri: str, description: str):
        """
        Initializes the object

        :param id: the id of the vuln
        :param osvdb_id: the osvdb_id of the vuln
        :param method: the method of the vuln
        :param iplink: the iplink of the vuln
        :param namelink: the namelink of the vuln
        :param uri: the uri of the vuln
        :param description: the description of the vuln
        """
        self.id = id
        self.osvdb_id = osvdb_id
        self.method = method
        self.iplink = iplink
        self.namelink = namelink
        self.uri = uri
        self.description = description

    def to_obs(self) -> EmulationVulnerabilityObservationState:
        """
        Converts the object into a VulnerabilityObservationState

        :return: the created VulnerabilityObservationState object
        """
        vuln = EmulationVulnerabilityObservationState(name="nikto_" + str(self.osvdb_id), port=None,
                                                      protocol=TransportProtocol.TCP,
                                                      cvss=0, osvdbid=self.osvdb_id, description=self.description,
                                                      service="http")
        return vuln

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"id:{self.id}, osvdb_id:{self.osvdb_id}, method:{self.method}, iplink:{self.iplink}, " \
               f"namelink:{self.namelink}, uri:{self.uri}, descr:{self.description}"



from typing import List, Dict, Any
from csle_common.dao.emulation_config.client_population_process_type import ClientPopulationProcessType
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_common.util.general_util import GeneralUtil


class ClientPopulationConfig:
    """
    A DTO object representing the configuration of the client population of an emulation
    """

    def __init__(self, ip: str, networks: List[ContainerNetwork], client_process_type: ClientPopulationProcessType,
                 lamb: float, mu: float, client_manager_port: int, client_manager_log_file: str,
                 client_manager_log_dir: str, client_manager_max_workers: int,
                 num_commands: int = 5,
                 client_time_step_len_seconds: int = 1, time_scaling_factor: float = 0.01,
                 period_scaling_factor: float = 20, docker_gw_bridge_ip: str = "", physical_host_ip: str = ""):
        """
        Creates a ClientPopulationConfig DTO Object

        :param ip: the ip of the client container
        :param networks: a list of networks in the emulation that are accessible for external networks
        :param client_process_type: the type of client arrival process (e.g. a Poisson process)
        :param lamb: the lambda parameter of the arrival process
        :param mu: the service-time parameter of the arrivals
        :param client_time_step_len_seconds: time-step length to measure the arrival process
        :param time_scaling_factor: the time-scaling factor for sine-modulated arrival processes
        :param period_scaling_factor: the period-scaling factor for sine-modulated arrival processes
        :param client_manager_log_file: the log file of the client manager
        :param client_manager_log_dir: the log dir of the client manager
        :param client_manager_max_workers: the maximum number of GRPC workers for the client manager
        :param docker_gw_bridge_ip: IP to reach the container from the host network
        :param physical_host_ip: IP of the physical host where the container is running
        """
        self.networks = networks
        self.ip = ip
        self.client_process_type = client_process_type
        self.lamb = lamb
        self.mu = mu
        self.client_manager_port = client_manager_port
        self.num_commands = num_commands
        self.client_time_step_len_seconds = client_time_step_len_seconds
        self.time_scaling_factor = time_scaling_factor
        self.period_scaling_factor = period_scaling_factor
        self.client_manager_log_dir = client_manager_log_dir
        self.client_manager_log_file = client_manager_log_file
        self.client_manager_max_workers = client_manager_max_workers
        self.docker_gw_bridge_ip = docker_gw_bridge_ip
        self.physical_host_ip = physical_host_ip

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ClientPopulationConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = ClientPopulationConfig(
            ip=d["ip"],
            networks=list(map(lambda x: ContainerNetwork.from_dict(x), d["networks"])),
            client_process_type=d["client_process_type"],
            lamb=d["lamb"], mu=d["mu"], client_manager_port=d["client_manager_port"],
            num_commands=d["num_commands"], client_time_step_len_seconds=d["client_time_step_len_seconds"],
            period_scaling_factor=d["period_scaling_factor"], time_scaling_factor=d["time_scaling_factor"],
            client_manager_log_dir=d["client_manager_log_dir"], client_manager_log_file=d["client_manager_log_file"],
            client_manager_max_workers=d["client_manager_max_workers"],
            docker_gw_bridge_ip=d["docker_gw_bridge_ip"], physical_host_ip=d["physical_host_ip"]
        )
        return obj

    def no_clients(self) -> "ClientPopulationConfig":
        """
        :return: A version of the config with no clients
        """
        return ClientPopulationConfig(
            ip=self.ip,
            networks=self.networks,
            client_process_type=self.client_process_type,
            lamb=0, mu=0, client_manager_port=self.client_manager_port,
            num_commands=0, client_time_step_len_seconds=self.client_time_step_len_seconds,
            period_scaling_factor=self.period_scaling_factor, time_scaling_factor=self.time_scaling_factor,
            client_manager_log_file="client_manager.log", client_manager_log_dir="/", client_manager_max_workers=10,
            docker_gw_bridge_ip=self.docker_gw_bridge_ip, physical_host_ip=self.physical_host_ip
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the object
        """
        d = {}
        d["ip"] = self.ip
        d["client_process_type"] = self.client_process_type
        d["lamb"] = self.lamb
        d["mu"] = self.mu
        d["networks"] = list(map(lambda x: x.to_dict(), self.networks))
        d["num_commands"] = self.num_commands
        d["client_manager_port"] = self.client_manager_port
        d["client_time_step_len_seconds"] = self.client_time_step_len_seconds
        d["time_scaling_factor"] = self.time_scaling_factor
        d["period_scaling_factor"] = self.period_scaling_factor
        d["client_manager_log_file"] = self.client_manager_log_file
        d["client_manager_log_dir"] = self.client_manager_log_dir
        d["client_manager_max_workers"] = self.client_manager_max_workers
        d["docker_gw_bridge_ip"] = self.docker_gw_bridge_ip
        d["physical_host_ip"] = self.physical_host_ip
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ip:{self.ip}, client_population_process_type: {self.client_process_type}, lamb:{self.lamb}, " \
               f"mu:{self.mu}, self.networks:{list(map(lambda x: str(x), self.networks))}, " \
               f"client_manager_port: {self.client_manager_port}, num_commands:{self.num_commands}, " \
               f"client_time_step_len_seconds: {self.client_time_step_len_seconds}," \
               f"time_scaling_factor: {self.time_scaling_factor}, " \
               f"period_scaling_factor: {self.period_scaling_factor}," \
               f"client_manager_log_file: {self.client_manager_log_file}, " \
               f"client_manager_log_dir: {self.client_manager_log_dir}, " \
               f"client_manager_max_workers: {self.client_manager_max_workers}, " \
               f"docker_gw_bridge_ip:{self.docker_gw_bridge_ip}, physical_host_ip: {self.physical_host_ip}"

    def to_json_str(self) -> str:
        """
        Converts the DTO into a json string

        :return: the json string representation of the DTO
        """
        import json
        json_str = json.dumps(self.to_dict(), indent=4, sort_keys=True)
        return json_str

    def to_json_file(self, json_file_path: str) -> None:
        """
        Saves the DTO to a json file

        :param json_file_path: the json file path to save  the DTO to
        :return: None
        """
        import io
        json_str = self.to_json_str()
        with io.open(json_file_path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    @staticmethod
    def from_json_file(json_file_path: str) -> "ClientPopulationConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ClientPopulationConfig.from_dict(json.loads(json_str))

    def copy(self) -> "ClientPopulationConfig":
        """
        :return: a copy of the DTO
        """
        return ClientPopulationConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "ClientPopulationConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.ip = GeneralUtil.replace_first_octet_of_ip(ip=config.ip, ip_first_octet=ip_first_octet)
        config.networks = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet),
                                   config.networks))
        return config



from typing import Dict, Any
from abc import ABC, abstractmethod


class SimulationEnvInputConfig(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass

    @staticmethod
    @abstractmethod
    def from_dict(d: Dict[str, Any]) -> "SimulationEnvInputConfig":
        pass




class AgentLog:
    """
    DTO Representing the Agent Log
    """

    def __init__(self):
        """
        Initializes the log
        """
        self.log = []

    def add_entry(self, msg) -> None:
        """
        Adds an entry to the log

        :param msg: the msg to add to the log
        :return: None
        """
        self.log.append(msg)

    def reset(self) -> None:
        """
        Resets the log

        :return: None
        """
        self.log = []



from typing import List
from csle_common.dao.emulation_observation.common.emulation_vulnerability_observation_state import \
    EmulationVulnerabilityObservationState
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.credential import Credential


class NmapVuln:
    """
    DTO representing a vulnerability found with NMAP
    """

    def __init__(self, name: str, port: int, protocol: TransportProtocol, cvss: float, service: str,
                 credentials: List[Credential] = None):
        """
        Initializes the DTO

        :param name: the name of the vulnerability
        :param port: the port of the vulnerability
        :param protocol: the protocol of the vulnerability
        :param cvss: the cvss of the vulnerability
        :param service: the service of the vulnerability
        :param credentials: the credentials of the vulnerability
        """
        self.name = name
        self.port = port
        self.protocol = protocol
        self.cvss = cvss
        self.service = service
        self.credentials = credentials
        if self.credentials is None:
            self.credentials = []

    def to_obs(self) -> EmulationVulnerabilityObservationState:
        """
        Converts the object into a VulnerabilityObservationState

        :return: the created VulnerabilityObservationState
        """
        service = ""
        if len(self.credentials) > 0:
            service = self.credentials[0].service
        vuln = EmulationVulnerabilityObservationState(name=self.name, port=self.port, protocol=self.protocol,
                                                      cvss=self.cvss, credentials=self.credentials, service=service)
        return vuln

    def __hash__(self) -> int:
        """
        :return: a hash representation of the object
        """
        return hash(self.name) + 31 * hash(self.port)

    def __eq__(self, other):
        """
        Compares equality of the object with another object

        :param other: the object to compare with
        :return: True if equal otherwise False
        """
        return (self.name == other.name and
                self.port == other.kafka_port)

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"name:{self.name}, port:{self.port}, protocol:{self.protocol}, cvss:{self.cvss}, " \
               f"service:{self.service}, credentials:{list(map(lambda x: str(x), self.credentials))}"



from typing import List, Dict, Union
import numpy as np
import torch
from stable_baselines3 import DQN
from csle_common.dao.training.policy import Policy
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.simulation_config.state import State
from csle_common.dao.simulation_config.state_type import StateType
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.logging.log import Logger


class DQNPolicy(Policy):
    """
    A neural network policy learned with DQN
    """

    def __init__(self, model, simulation_name: str, save_path: str, player_type: PlayerType, states: List[State],
                 actions: List[Action], experiment_config: ExperimentConfig, avg_R: float):
        """
        Initializes the policy

        :param model: the DQN model
        :param simulation_name: the simulation name
        :param save_path: the path to save the model to
        :param states: list of states (required for computing stage policies)
        :param actions: list of actions
        :param experiment_config: the experiment configuration for training the policy
        :param avg_R: the average reward of the policy when evaluated in the simulation
        """
        super(DQNPolicy, self).__init__(agent_type=AgentType.DQN, player_type=player_type)
        self.model = model
        self.id = -1
        self.simulation_name = simulation_name
        self.save_path = save_path
        if self.model is None:
            try:
                self.model = DQN.load(path=self.save_path)
            except Exception as e:
                Logger.__call__().get_logger().warning(
                    f"There was an exception loading the model from path: {self.save_path}, "
                    f"exception: {str(e)}, {repr(e)}")
        self.states = states
        self.actions = actions
        self.experiment_config = experiment_config
        self.avg_R = avg_R

    def action(self, o: List[float]) -> Union[int, List[int], np.ndarray]:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :return: the selected action
        """
        a, _ = self.model.predict(np.array(o), deterministic=False)
        return a

    def probability(self, o: List[float], a) -> Union[int, List[int], np.ndarray]:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :return: the selected action
        """
        actions = self.model.policy.forward(obs=torch.tensor(o).to(self.model.device))
        action = actions[0]
        if action == a:
            return 1
        else:
            return 0

    def to_dict(self) -> Dict[str, Union[float, int, str]]:
        """
        :return: a dict representation of the policy
        """
        d = {}
        d["id"] = self.id
        d["simulation_name"] = self.simulation_name
        d["save_path"] = self.save_path
        if self.model is not None:
            d["policy_kwargs"] = self.model.policy_kwargs
            self.model.save(path=self.save_path)
        else:
            d["policy_kwargs"] = {}
            d["policy_kwargs"]["net_arch"] = []
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        d["player_type"] = self.player_type
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["experiment_config"] = self.experiment_config.to_dict()
        d["agent_type"] = self.agent_type
        d["avg_R"] = self.avg_R
        return d

    @staticmethod
    def from_dict(d: Dict) -> "DQNPolicy":
        """
        Converst a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = DQNPolicy(model=None, simulation_name=d["simulation_name"], save_path=d["save_path"],
                        states=list(map(lambda x: State.from_dict(x), d["states"])), player_type=d["player_type"],
                        actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
                        experiment_config=ExperimentConfig.from_dict(d["experiment_config"]), avg_R=d["avg_R"])
        obj.id = d["id"]
        return obj

    def stage_policy(self, o: Union[List[Union[int, float]], int, float]) -> List[List[float]]:
        """
        Gets the stage policy, i.e a |S|x|A| policy

        :param o: the latest observation
        :return: the |S|x|A| stage policy
        """
        b1 = o[1]
        l = int(o[0])
        if not self.player_type == PlayerType.ATTACKER:
            stage_policy = []
            for _ in self.states:
                stage_policy.append(self._get_attacker_stopping_dist(obs=o))
            return stage_policy
        else:
            stage_policy = []
            for s in self.states:
                if s.state_type != StateType.TERMINAL:
                    o = [l, b1, s.id]
                    stage_policy.append(self._get_attacker_stopping_dist(obs=o))
                else:
                    stage_policy.append([0.5, 0.5])
            return stage_policy

    def _get_attacker_stopping_dist(self, obs) -> List[float]:
        """
        Utility function for getting the stopping action distribution conditioned on a given observation

        :param obs: the observation to condition on
        :return: the conditional ation distribution
        """
        obs = np.array([obs])
        actions = self.model.policy.forward(obs=torch.tensor(obs).to(self.model.device))
        action = actions[0]
        if action == 1:
            stop_prob = 1
        else:
            stop_prob = 0
        return [1 - stop_prob, stop_prob]

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"model: {self.model}, id: {self.id}, simulation_name: {self.simulation_name}, " \
               f"save path: {self.save_path}, states: {self.states}, experiment_config: {self.experiment_config}," \
               f"avg_R: {self.avg_R}"

    def to_json_str(self) -> str:
        """
        Converts the DTO into a json string

        :return: the json string representation of the DTO
        """
        import json
        json_str = json.dumps(self.to_dict(), indent=4, sort_keys=True)
        return json_str

    def to_json_file(self, json_file_path: str) -> None:
        """
        Saves the DTO to a json file

        :param json_file_path: the json file path to save  the DTO to
        :return: None
        """
        import io
        json_str = self.to_json_str()
        with io.open(json_file_path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    @staticmethod
    def from_json_file(json_file_path: str) -> "DQNPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return DQNPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "DQNPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())



from typing import Tuple
import socket


class GeneralUtil:
    """
    Class with general utility functions
    """

    @staticmethod
    def get_host_ip() -> str:
        """
        Utility method for getting the ip of the host

        :return: the ip of the host
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        return s.getsockname()[0]

    @staticmethod
    def replace_first_octet_of_ip(ip: str, ip_first_octet: int) -> str:
        """
        Utility function for changing the first octet in an IP address

        :param ip: the IP to modify
        :param ip_first_octet: the first octet to insert
        :return: the new IP
        """
        index_of_first_octet_end = ip.find(".")
        return str(ip_first_octet) + ip[index_of_first_octet_end:]

    @staticmethod
    def replace_first_octet_of_ip_tuple(tuple_of_ips: Tuple[str, str], ip_first_octet: int) -> Tuple[str, str]:
        """
        Utility function for changing the first octet in an IP address

        :param ip: the IP to modify
        :param ip_first_octet: the first octet to insert
        :return: the new IP
        """
        index_of_first_octet_end = tuple_of_ips[0].find(".")
        first_ip = str(ip_first_octet) + tuple_of_ips[0][index_of_first_octet_end:]
        index_of_first_octet_end = tuple_of_ips[1].find(".")
        second_ip = str(ip_first_octet) + tuple_of_ips[1][index_of_first_octet_end:]
        return (first_ip, second_ip)




class NmapOs:
    """
    DTO representing the operating system found with an NMAP scan
    """

    def __init__(self, name: str, vendor: str, osfamily: str, accuracy: int):
        """
        Initializes the DTO object

        :param name: the name of the operating system
        :param vendor: the vendor of the operating system
        :param osfamily: the family of the operating system
        :param accuracy: the accuracy of the OS guess
        """
        self.name = name
        self.vendor = vendor
        self.osfamily = osfamily
        self.accuracy = accuracy

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"name:{self.name}, vendor:{self.vendor}, os_family:{self.osfamily}, accuracy:{self.accuracy}"

    @staticmethod
    def get_best_match(os_matches):
        """
        Returns the best matching os

        :param os_matches: list of os matches
        :return: best matching os
        """
        best_accuracy = 0
        best_os = None
        for os in os_matches:
            if os.accuracy > best_accuracy:
                best_os = os
        return best_os



from typing import List, Dict, Any
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.credential import Credential


class NetworkService:
    """
    DTO Class representing a serice in the network
    """

    def __init__(self, protocol: TransportProtocol, port: int, name: str, credentials: List[Credential] = None):
        """
        Initializes the DTO

        :param protocol: the protocol of the service
        :param port: the port of the service
        :param name: the name of the service
        :param credentials: the list of credentials of the service
        """
        self.protocol = protocol
        self.port = port
        self.name = name
        self.credentials = credentials

    def __str__(self) -> str:
        """
        :return: a string representation of the service
        """
        cr = []
        if self.credentials is not None:
            cr = list(map(lambda x: str(x), self.credentials))
        return "protocol:{}, port:{}, name:{}, credentials: {}".format(self.protocol, self.port, self.name,
                                                                       cr)

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the object
        """
        d = {}
        d["protocol"] = self.protocol
        d["port"] = self.port
        d["name"] = self.name
        d["credentials"] = list(map(lambda x: x.to_dict(), self.credentials))
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NetworkService":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = NetworkService(name=d["name"], port=d["port"], protocol=d["protocol"],
                             credentials=list(map(lambda x: Credential.from_dict(x), d["credentials"])))
        return dto

    def copy(self) -> "NetworkService":
        """
        :return: a copy of the DTO
        """
        return NetworkService(
            protocol=self.protocol, port=self.port, name=self.name, credentials=self.credentials
        )

    @staticmethod
    def pw_vuln_services():
        """
        :return: a list of all vulnerabilities that involve weak passwords
        """
        ssh_vuln_service = (NetworkService(protocol=TransportProtocol.TCP, port=22, name="ssh", credentials=[]),
                            constants.EXPLOIT_VULNERABILITES.SSH_DICT_SAME_USER_PASS)
        ftp_vuln_service = (NetworkService(protocol=TransportProtocol.TCP, port=21, name="ftp", credentials=[]),
                            constants.EXPLOIT_VULNERABILITES.FTP_DICT_SAME_USER_PASS)
        telnet_vuln_service = (NetworkService(protocol=TransportProtocol.TCP, port=23, name="telnet", credentials=[]),
                               constants.EXPLOIT_VULNERABILITES.TELNET_DICTS_SAME_USER_PASS)
        return [ssh_vuln_service, ftp_vuln_service, telnet_vuln_service], [ssh_vuln_service, telnet_vuln_service]

    def __eq__(self, other) -> bool:
        """
        Tests equality with another service

        :param other: the service to compare with
        :return: True if equal otherwise False
        """
        if not isinstance(other, NetworkService):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.name == other.name and self.protocol == other.protocol and self.port == other.port

    @staticmethod
    def from_credential(credential: Credential) -> "NetworkService":
        """
        Converts the object into a network service representation

        :return: the network service representation
        """
        service = NetworkService(protocol=credential.protocol, port=credential.port,
                                 name=credential.service,
                                 credentials=[credential])
        return service

    def to_json_str(self) -> str:
        """
        Converts the DTO into a json string

        :return: the json string representation of the DTO
        """
        import json
        json_str = json.dumps(self.to_dict(), indent=4, sort_keys=True)
        return json_str

    def to_json_file(self, json_file_path: str) -> None:
        """
        Saves the DTO to a json file

        :param json_file_path: the json file path to save  the DTO to
        :return: None
        """
        import io
        json_str = self.to_json_str()
        with io.open(json_file_path, 'w', encoding='utf-8') as f:
            f.write(json_str)

    @staticmethod
    def from_json_file(json_file_path: str) -> "NetworkService":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NetworkService.from_dict(json.loads(json_str))



from typing import List
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state \
    import EmulationConnectionObservationState
from csle_common.dao.emulation_config.credential import Credential
from csle_common.tunneling.forward_tunnel_thread import ForwardTunnelThread


class ConnectionSetupDTO:
    """
    DTO class containing information for setting up connections in the emulation
    """

    def __init__(self, connected: bool = False, credentials: List[Credential] = None,
                 target_connections: List = None,
                 tunnel_threads: List[ForwardTunnelThread] = None, forward_ports: List[int] = None,
                 ports: List[int] = None, interactive_shells: List = None, total_time: float = 0.0,
                 non_failed_credentials: List[Credential] = None,
                 proxies: List[EmulationConnectionObservationState] = None, ip: str = None):
        """
        Initializes the DTO

        :param connected: whether the connection is connected or not
        :param credentials: the list of credentials
        :param target_connections: the list of target connections
        :param tunnel_threads: the list of tunnel threads
        :param forward_ports: the list of forward ports
        :param ports: the list of ports
        :param interactive_shells: the list of interactive shells
        :param total_time: the total time of connection
        :param non_failed_credentials: the non-failed-credentials
        :param proxies: the list of proxy connections
        :param ip: the ip address of the connectio
        """
        self.connected = connected
        self.total_time = total_time
        self.credentials = credentials
        self.target_connections = target_connections
        self.tunnel_threads = tunnel_threads
        self.forward_ports = forward_ports
        self.ports = ports
        self.interactive_shells = interactive_shells
        self.total_time = total_time
        self.non_failed_credentials = non_failed_credentials
        self.proxies = proxies
        self.ip = ip

        if self.target_connections is None:
            self.target_connections = []
        if self.tunnel_threads is None:
            self.tunnel_threads = []
        if self.credentials is None:
            self.credentials = []
        if self.forward_ports is None:
            self.forward_ports = []
        if self.ports is None:
            self.ports = []
        if self.interactive_shells is None:
            self.interactive_shells = []
        if self.non_failed_credentials is None:
            self.non_failed_credentials = []
        if self.proxies is None:
            self.proxies = []

    def __str__(self) -> str:
        """
        :return: a string represetation of the object
        """
        return f"connected:{self.connected}, total_time:{self.total_time}, " \
               f"credentials:{list(map(lambda x: str(x), self.credentials))}, " \
               f"target_connections:{list(map(lambda x: str(x), self.target_connections))}, " \
               f"tunnel_threads:{list(map(lambda x: str(x), self.tunnel_threads))}, " \
               f"forward_ports:{list(map(lambda x: str(x), self.forward_ports))}, " \
               f"ports:{list(map(lambda x: str(x), self.ports))}, " \
               f"interactive_shells:{list(map(lambda x: str(x), self.interactive_shells))}, " \
               f"non_failed_credentials:{list(map(lambda x: str(x), self.non_failed_credentials))}, " \
               f"proxies:{list(map(lambda x: str(x), self.proxies))}, ip:{self.ip}"

    def copy(self) -> "ConnectionSetupDTO":
        """
        :return: a copy of the object
        """
        return ConnectionSetupDTO(
            connected=self.connected, credentials=self.credentials, target_connections=self.target_connections,
            tunnel_threads=self.tunnel_threads, forward_ports=self.forward_ports, ports=self.ports,
            interactive_shells=self.interactive_shells, total_time=self.total_time,
            non_failed_credentials=self.non_failed_credentials, proxies=self.proxies, ip=self.ip
        )
